package com.goldencode.p2j.util;

import java.io.*;

import com.goldencode.util.*;

public abstract class Text
extends BaseDataType
{
   /** Charset name to use for encoding or <code>null</code> for default. */
   protected static final String override = Utils.getCharsetOverride();
   
   /**
    * Stores the state of this character as an instance of class
    * <code>String</code>.  This is independent of the state of whether or
    * not this instance represents the <code>unknown value</code>.
    */
   protected String value = null; 
   
   /**
    * Stores the state of whether or not this instance represents the
    * <code>unknown value</code>, independent of the state of this
    * character's data.
    */
   protected boolean unknown = false;
   
   /**
    * Stores the state of whether or not this instance compares case
    * sensitively.
    */
   protected boolean caseSens = false;
   
   /**
    * Default constructor, creates an instance that represents the unknown
    * value.
    */
   public Text()
   {
      unknown = true;
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a DYNAMIC-FUNCTION call to the expected type (i.e
    * when passing that value as a parameter to another function or procedure
    * call).  In such cases, 4GL does some automatic conversion, which is not
    * implemented yet.
    * <p>
    * This version of the c'tor always sets the value to unknown.
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public Text(BaseDataType value)
   {
      this();
      // no-op
   }
   
   /**
    * Constructs an instance that has the <code>String</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>,
    * unless the <code>value</code> argument is <code>null</code>.
    *
    * @param    value
    *           The character value of the resulting instance.
    */
   public Text(String value)
   {
      if (value == null)
      {
         setUnknown();
      }
      else
      {
         this.value = value;
      }
   }
   
   /**
    * Constructs an instance that has the <code>String</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>
    * unless the <code>value</code> argument is <code>null</code>.
    *
    * @param    value
    *           The character value of the resulting instance.
    * @param    caseSens
    *           Specifies whether this instance should use case sensitive
    *           comparisons.
    */
   public Text(String value, boolean caseSens)
   {
      if (value == null)
      {
         setUnknown();
      }
      else
      {
         this.value    = value;
         this.caseSens = caseSens;
      }
   }
   
   /**
    * Constructs an instance that has a <code>char</code> value passed
    * in and which <b>does not</b> represent the <code>unknown value</code>. 
    *
    * @param    ch
    *           The character value of the resulting instance.
    */
   public Text(char ch)
   {
      this.value = new StringBuffer().append(ch).toString();
   }
   
   /**
    * Create a new instance, by copying the state of the passed {@link Text}.
    * If the <code>output</code> parameter is set to <code>true</code>, it 
    * means that the referenced value is saved, as the passed instance will
    * be passed to a function call, and the state might need to be updated
    * on function exit. 
    * 
    * @param    value
    *           The value which needs to be copied.
    * @param    output
    *           Flag indicating if this instance is passed to as a reference
    *           for an output parameter.
    */
   public Text(Text value, boolean output)
   {
      deepAssign(value);

      if (output)
      {
         // TODO: save a reference to the variable
      }
   }

   /**
    * Returns the larger (lexicographically) of the current or passed-in
    * instance.
    *
    * @param    c
    *           The instance to compare against.
    *
    * @return   The largest instance.
    */
   public Text maximum(Text c)   
   {
      return (Text) super.maximum(c);
   }

   /**
    * Returns the smaller (lexicographically) of the current or passed-in 
    * instance.
    *
    * @param    c
    *           The instance to compare against.
    *
    * @return   The smallest instance.
    */
   public Text minimum(Text c)   
   {
      return (Text) super.minimum(c);
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, double pos)
   throws ErrorConditionException
   {
      replaceWorker(text, new integer(pos), new integer(-1), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, double pos, double len)
   throws ErrorConditionException
   {
      replaceWorker(text, new integer(pos), new integer(len), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    * @param    type
    *           Description of how to interpret the position and length values.
    *           At this time only "character" is supported.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, double pos, double len, String type)
   throws ErrorConditionException
   {
      checkOverlayType(type);
      replaceWorker(text, new integer(pos), new integer(len), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, double pos, NumberType len)
   throws ErrorConditionException
   {
      replaceWorker(text, new integer(pos), len, true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    * @param    type
    *           Description of how to interpret the position and length values.
    *           At this time only "character" is supported.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text  text, double pos, NumberType len, String type)
   throws ErrorConditionException
   {
      checkOverlayType(type);
      replaceWorker(text, new integer(pos), len, true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, NumberType pos)
   throws ErrorConditionException
   {
      replaceWorker(text, pos, new integer(-1), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, NumberType pos, double len)
   throws ErrorConditionException
   {
      replaceWorker(text, pos, new integer(len), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    * @param    type
    *           Description of how to interpret the position and length values.
    *           At this time only "character" is supported.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, NumberType pos, double len, String type)
   throws ErrorConditionException
   {
      checkOverlayType(type);
      replaceWorker(text, pos, new integer(len), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, NumberType pos, NumberType len)
   throws ErrorConditionException
   {
      replaceWorker(text, pos, len, true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    * @param    type
    *           Description of how to interpret the position and length values.
    *           At this time only "character" is supported.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(Text text, NumberType pos, NumberType len, String type)
   throws ErrorConditionException
   {
      checkOverlayType(type);
      replaceWorker(text, pos, len, true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, double pos)
   throws ErrorConditionException
   {
      replaceWorker(new character(text),
                    new integer(pos),
                    new integer(-1),
                    true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, double pos, double len)
   throws ErrorConditionException
   {
      replaceWorker(new character(text),
                    new integer(pos),
                    new integer(len),
                    true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    * @param    type
    *           Description of how to interpret the position and length values.
    *           At this time only "character" is supported.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, double pos, double len, String type)
   throws ErrorConditionException
   {
      checkOverlayType(type);
      replaceWorker(new character(text),
                    new integer(pos),
                    new integer(len),
                    true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, double pos, NumberType len)
   throws ErrorConditionException
   {
      replaceWorker(new character(text), new integer(pos), len, true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    * @param    type
    *           Description of how to interpret the position and length values.
    *           At this time only "character" is supported.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, double pos, NumberType len, String type)
   throws ErrorConditionException
   {
      checkOverlayType(type);
      replaceWorker(new character(text), new integer(pos), len, true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, NumberType pos)
   throws ErrorConditionException
   {
      replaceWorker(new character(text), pos, new integer(-1), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, NumberType pos, double len)
   throws ErrorConditionException
   {
      replaceWorker(new character(text), pos, new integer(len), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    * @param    type
    *           Description of how to interpret the position and length values.
    *           At this time only "character" is supported.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, NumberType pos, double len, String type)
   throws ErrorConditionException
   {
      checkOverlayType(type);
      replaceWorker(new character(text), pos, new integer(len), true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String text, NumberType pos, NumberType len)
   throws ErrorConditionException
   {
      replaceWorker(new character(text), pos, len, true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace
    *           and the number of characters from the replacement text to 
    *           include in the replacement.
    * @param    type
    *           Description of how to interpret the position and length values.
    *           At this time only "character" is supported.
    *
    * @return   The current instance is returned.
    */
   public Text overlay(String     text,
                       NumberType pos,
                       NumberType len,
                       String     type)
   throws ErrorConditionException
   {
      checkOverlayType(type);
      replaceWorker(new character(text), pos, len, true);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    *
    * @return   The current instance is returned.
    */
   public Text replace(Text text, double pos)
   throws ErrorConditionException
   {
      replaceWorker(text, new integer(pos), new integer(-1), false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    *
    * @return   The current instance is returned.
    */
   public Text replace(Text text, double pos, double len)
   throws ErrorConditionException
   {
      replaceWorker(text, new integer(pos), new integer(len), false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    *
    * @return   The current instance is returned.
    */
   public Text replace(Text text, double pos, NumberType len)
   throws ErrorConditionException
   {
      replaceWorker(text, new integer(pos), len, false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    *
    * @return   The current instance is returned.
    */
   public Text replace(Text text, NumberType pos)
   throws ErrorConditionException
   {
      replaceWorker(text, pos, new integer(-1), false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    *
    * @return   The current instance is returned.
    */
   public Text replace(Text text, NumberType pos, double len)
   throws ErrorConditionException
   {
      replaceWorker(text, pos, new integer(len), false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    *
    * @return   The current instance is returned.
    */
   public Text replace(Text text, NumberType pos, NumberType len)
   throws ErrorConditionException
   {
      replaceWorker(text, pos, len, false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    *
    * @return   The current instance is returned.
    */
   public Text replace(String text, double pos)
   throws ErrorConditionException
   {
      replaceWorker(new character(text),
                    new integer(pos),
                    new integer(-1),
                    false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    *
    * @return   The current instance is returned.
    */
   public Text replace(String text, double pos, double len)
   throws ErrorConditionException
   {
      replaceWorker(new character(text),
                    new integer(pos),
                    new integer(len),
                    false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    *
    * @return   The current instance is returned.
    */
   public Text replace(String text, double pos, NumberType len)
   throws ErrorConditionException
   {
      replaceWorker(new character(text), new integer(pos), len, false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    *
    * @return   The current instance is returned.
    */
   public Text replace(String text, NumberType pos)
   throws ErrorConditionException
   {
      replaceWorker(new character(text), pos, new integer(-1), false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    *
    * @return   The current instance is returned.
    */
   public Text replace(String text, NumberType pos, double len)
   throws ErrorConditionException
   {
      replaceWorker(new character(text), pos, new integer(len), false);
      return this;
   }

   /**
    * Overwrites characters in the current value of this instance with the 
    * passed text at the 1-based index into the string.  The entire
    * replacement text will be placed into the resulting value.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    *
    * @return   The current instance is returned.
    */
   public Text replace(String text, NumberType pos, NumberType len)
   throws ErrorConditionException
   {
      replaceWorker(new character(text), pos, len, false);
      return this;
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    ii
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.
    */
   public Text replaceEntry(double ii, char delimit, String replace)
   throws ErrorConditionException
   {
      boolean content  = (value.length() > 0); 
      boolean endDelim = content && 
                         (value.charAt(value.length() - 1) == delimit); 
      
      // create the regular expression (with proper escapes)
      String regex = TextOps.fixupRegexChar(delimit);      
      
      // split up the list (make sure all possible entries are included,
      // even empty trailing entries --> use -1 as the limit to force this) 
      String[] array = value.split(regex, -1);
      
      // convert 1 based index to 0 based
      int i   = (int) ii;
      int idx = i - 1;
      
      if (idx < 0 || idx >= array.length)
      {
         TextOps.genOutsideRangeError(i, value);
         return this;
      }
      
      StringBuffer sb = new StringBuffer();
      
      // put the list back together
      for (int j = 0; j < array.length; j++)
      {
         // add the delimiter on all subsequent iterations
         if (j > 0)
         {
            sb.append(delimit);
         }
         
         if (j == idx)
         {
            // replace this element
            sb.append(replace);
         }
         else
         {
            // emit this element (even if empty)
            sb.append(array[j]);
         }
      }
      
      // add an ending delimiter if there was one to start
      if (endDelim)
      {
         sb.append(delimit);
      }
      
      // save the result
      value = sb.toString();
      
      return this;
   }

   /**
    * Replaces the specified 1-based indexed comma-delimited entry in the
    * current instance which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(double i, Text replace)
   throws ErrorConditionException
   {
      if (replace.unknown)
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry((int) i, ',', replace.value);
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(double i, character delimit, Text replace)
   throws ErrorConditionException
   {
      if (replace.unknown || delimit.unknown)
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry((int) i,
                          TextOps.safeDelimiter(delimit.value),
                          replace.value);
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(double i, character delimit, String replace)
   throws ErrorConditionException
   {
      if (delimit.unknown)
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry((int) i, 
                          TextOps.safeDelimiter(delimit.value),
                          replace);
   }

   /**
    * Replaces the specified 1-based indexed comma-delimited entry in the
    * current instance which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.
    */
   public Text replaceEntry(double i, String replace)
   throws ErrorConditionException
   {
      return replaceEntry((int) i, ',', replace);
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(double i, String delimit, Text replace)
   throws ErrorConditionException
   {
      if (replace.unknown)
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry((int) i,
                          TextOps.safeDelimiter(delimit),
                          replace.value);
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.
    */
   public Text replaceEntry(double i, String delimit, String replace)
   throws ErrorConditionException
   {
      return replaceEntry((int) i, TextOps.safeDelimiter(delimit), replace);
   }

   /**
    * Replaces the specified 1-based indexed comma-delimited entry in the
    * current instance which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(NumberType i, character replace)
   throws ErrorConditionException
   {
      if (i.isUnknown() || replace.unknown)
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry(i.intValue(), ',', replace.value);
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(NumberType i, character delimit, Text replace)
   throws ErrorConditionException
   {
      if (i.isUnknown() || replace.unknown || delimit.unknown)
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry(i.intValue(),
                          TextOps.safeDelimiter(delimit.value),
                          replace.value);
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(NumberType i, character delimit, String replace)
   throws ErrorConditionException
   {
      if (i.isUnknown() || delimit.unknown)
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry(i.intValue(),
                          TextOps.safeDelimiter(delimit.value),
                          replace);
   }

   /**
    * Replaces the specified 1-based indexed comma-delimited entry in the
    * current instance which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(NumberType i, String replace)
   throws ErrorConditionException
   {
      if (i.isUnknown())
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry(i.intValue(), ',', replace);
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(NumberType i, String delimit, Text replace)
   throws ErrorConditionException
   {
      if (i.isUnknown() || replace.unknown)
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry(i.intValue(),
                          TextOps.safeDelimiter(delimit),
                          replace.value);
   }

   /**
    * Replaces the specified 1-based indexed entry in the current instance
    * which must be a delimited list.  
    *
    * @param    i
    *           The 1-based index into the list.
    * @param    delimit
    *           The delimiter character.
    * @param    replace
    *           The replacment text.
    *
    * @return   The current instance.  It may have been set to the 
    *           <code>unknown value</code> if any of the inputs are
    *           <code>unknown value</code>.
    */
   public Text replaceEntry(NumberType i, String delimit, String replace)
   throws ErrorConditionException
   {
      if (i.isUnknown())
      {
         setUnknown();
         return this;
      }
      
      return replaceEntry(i.intValue(), 
                          TextOps.safeDelimiter(delimit),
                          replace);
   }

   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()   
   {
      return new character();
   }

   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(BaseDataType value, boolean force)
   {
      assign((Text) value, force);
   }

   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance. The configuration (case-sensitivity) of
    * this instance is not set.  Use {@link #deepAssign} if both state and
    * configuration must be set.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Text value)
   {
      assign(value, false);
   }

   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(Text value, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         this.value = value.value;
         unknown    = value.unknown;
      }
   }

   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance. The configuration (case-sensitivity) of
    * this instance is not set.  Use {@link #deepAssign} if both state and
    * configuration must be set.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state. If <code>null</code>
    *           is passed, the instance will be set to the unknown value.
    */
   public void assign(String value)
   {
      assign(new character(value), false);
   }

   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((Text) value, false);
   }

   /**
    * Sets the state (data and unknown value) and configuration 
    * (case-sensitivity) of this instance based on the state and configuration
    * of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state and configuration.
    */
   public void deepAssign(Text value)
   {
      assign(value, true);
      caseSens = value.caseSens;
   }

   /**
    * Accesses the state of this instance's <code>String</code> value
    * independently of the state of whether this instance represents the
    * <code>unknown value</code>.  If this instance represents the
    * <code>unknown value</code>, the returned <code>String</code> is
    * <b>not valid</b>! 
    *
    * @return   The state of this instance's <code>String</code> value.
    */
   public String getValue()
   {
      return value;
   }

   /**
    * Sets the state of this instance's <code>String</code> value
    * and sets the <code>unknown value</code> state for this instance to
    * <code>false</code>.
    *
    * @param    value
    *           The state to set this instance's <code>String</code> value.
    */
   public void setValue(String value)
   {
      this.value = value;
      unknown    = false;
   }

   /**
    * Determines if this instance compares case sensitively.
    *
    * @return   <code>true</code> if this instance is set to case sensitive.
    */
   public boolean isCaseSensitive()
   {
      return caseSens;
   }

   /**
    * Sets the state of this instance's compares case sensitive flag.
    * <p>
    * This method is used to set the current value in a manner that does not
    * schedule an automated reset to the prior value later. Thus, this method
    * is suitable for the initial configuration of a new instance.  If a
    * temporary case-sensitivity change is needed, use 
    * {@link #setTemporaryCaseSensitive}.
    *
    * @param    caseSens
    *           <code>true</code> if this instance must be set to case
    *           sensitive.
    */
   public void setCaseSensitive(boolean caseSens)
   {
      this.caseSens = caseSens;
   }

   /**
    * Sets the state of this instance's compares case sensitive flag for the
    * duration of the caller's current scope.  At the end of the caller's
    * scope a notification will trigger a reset of the case sensitive flag
    * to the current value.
    * <p>
    * This method should only be used when importing references via the
    * {@link SharedVariableManager} as a local instance or a new shared
    * instance would not need to be reset, since they would have no prior
    * value to which to return.  To set the initial case-sensitivity, please
    * se {@link #setCaseSensitive}.
    * <p>
    * The {@link TransactionManager#registerFinalizable} facility is used to
    * obtain the callback necessary to reset the sensitivity at end of the
    * caller's scope.
    *
    * @param    caseSens
    *           <code>true</code> if this instance must be set to case
    *           sensitive.
    */
   public void setTemporaryCaseSensitive(boolean caseSens)
   {
      // do nothing if the request precision is the same as the current
      // precision
      if (this.caseSens != caseSens)
      {
         // remember the old value
         boolean old = this.caseSens;
         
         // reset
         setCaseSensitive(caseSens);
         
         // schedule the reset
         SensitivityResetter sr = new SensitivityResetter(old);
         TransactionManager.registerFinalizable(sr, false);
      }
   }

   /**
    * Test equality between the internal values of this object and the given
    * parameter on a case sensitive basis.  Before the internal values are
    * checked, the unknown states of both operands are compared.  If both are
    * unknown, <code>true</code> is returned.  If only one is unknown,
    * <code>false</code> is returned.  Otherwise, the case sensitive test is
    * performed.
    *
    * @param   c
    *          Another character instance to test for case sensitive equality
    *          with this object.
    *
    * @return  <code>true</code> if this object and the given object test as
    *          equal, else <code>false</code>.
    */
   public boolean equalsCaseSensitively(Text c)
   {
      boolean u1 = isUnknown();
      boolean u2 = c.isUnknown(); 
      
      if (u1 && u2) 
         return true;
      
      if (u1 || u2)
         return false;
      
      // prepare the data
      String s1 = StringHelper.safeTrimTrailing(value);      
      String s2 = StringHelper.safeTrimTrailing(c.value);
      
      return s1.equals(s2);
   }

   /**
    * Hash code implementation which is consistent with {@link
    * #equalsCaseSensitively}.
    *
    * @return  Hash code value for this object instance.
    */
   public int caseSensitiveHashCode()
   {
      int result = 17;
      if (isUnknown())
      {
         return result;
      }
      
      result = 37 * result + value.hashCode();
      
      return result;
   }

   /**
    * Determines if this instance represents the <code>unknown value</code>.
    *
    * @return   <code>true</code> if this instance is set to the
    *           <code>unknown value</code>.
    */
   public boolean isUnknown()
   {
      return unknown;
   }

   /**
    * Sets the state of this instance's <code>unknown value</code> flag to
    * <code>true</code>.
    * <p>
    * <b>Warning: a separate call is needed to ensure that the data of this
    * instance is set to the correct value.</b>
    */
   public void setUnknown()
   {
      unknown = true;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      unknown  = in.readBoolean();
      caseSens = in.readBoolean();
      
      // the next object can be null so we read an indicator var first to
      // determine if there is following data to be read
      
      if (in.readBoolean())
         value = (String) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      out.writeBoolean(unknown);
      out.writeBoolean(caseSens);
      
      // the next object can be null so we write an indicator var first to
      // notify reader if there is following data to be read
      
      boolean there = (value != null); 
      out.writeBoolean(there);
      if (there)
         out.writeObject(value);
   }

   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * Core worker that handles comparisons between 2 string operands after
    * removing trailing whitespace and honoring the case sensitivity 
    * operands.  
    * <p>
    * Case-insensitive comparisons are obtained by manually uppercasing both
    * operands which is how Progress does it. Lowercase comparisons cannot be 
    * used becuase the results for {@link CompareOps#isGreaterThan},
    * {@link CompareOps#isLessThan}, {@link CompareOps#isGreaterThanOrEqual}
    * and {@link CompareOps#isLessThanOrEqual} would be different in Java
    * than in Progress when the following characters are included in the
    * string:
    * <pre>
    * [
    * \
    * ^
    * _
    * '
    * </pre>
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The instance to compare against.
    *
    * @return   -1, 0 or 1 depending on if this instance is less than,
    *           equal to or greater (respectively) than the specified
    *           instance <code>obj</code>.
    */
   public int compareTo(Object obj)   
   {
      Text c = null;
      
      if (obj instanceof Text)
      {
         c = (Text) obj;
      }
      else if (obj instanceof String)
      {
         c = new character((String) obj);
      }
      else
      {
         return -1;
      }
      
      // prepare the data
      String s1 = StringHelper.safeTrimTrailing(value);      
      String s2 = StringHelper.safeTrimTrailing(c.value);
      
      // if either instance requires case sensitivity, force a case
      // sensitive comparison
      if (caseSens || c.caseSens)
      {
         
         return s1.compareTo(s2);
      }
               
      // DO NOT use String.compareToIgnoreCase() since this lowercases
      // and yields different results for >, <, >= and <= forms when
      // [ \ ^ _ ' characters are included in the operands
      return s1.toUpperCase().compareTo(s2.toUpperCase());
   }

   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance.
    */
   public int hashCode()
   {
      int result = 17;
      if (isUnknown())
      {
         return result;
      }
      
      String s = (caseSens ? value : value.toUpperCase());
      result = 37 * result + s.hashCode();
      
      return result;
   }

   /**
    * Return the current value as a Java wrapper.
    *
    * @return  Current value or <code>null</code> if unknown.
    */
   public String toJavaType()
   {
      return (isUnknown() ? null : getValue());
   }

   /**
    * Check the overlay type description and throw an exception if the type
    * is not yet supported. The check is case-insensitive.
    *
    * @param    type
    *           Must be "character" at this time.
    *
    * @throws   IllegalArgumentException
    *           If the type doesn't match "character".
    */
   private void checkOverlayType(String type)
   {
      if (type != null && !type.equalsIgnoreCase("character"))
      {
         String err = String.format("OVERLAY type %s is not supported!", type);
         throw new IllegalArgumentException(err);
      }
   }

   /**
    * Core method to implement overlay and replace support. This method
    * changes the current value of this instance, overwriting characters
    * at the specified index given by <code>pos</code> with the text
    * specified by <code>text</code>.  The number of replaced characters
    * is specified by <code>len</code>, with a negative number indicating
    * that the length of characters in <code>text</code> is used.  If
    * <code>overlay</code> mode is requested, then the <code>len</code> 
    * parameter also limits the number of characters in the <code>text</code>
    * parameter which will be used as the replacement.  If any of the
    * parameters are <code>unknown value</code>, the instance will be set
    * to the <code>unknown value</code>.
    *
    * @param    text
    *           Replacement text.
    * @param    pos
    *           The 1-based index into the string at which the replacement
    *           is to occur.
    * @param    len
    *           The number of characters in the source string to replace.
    * @param    overlay
    *           <code>true</code> if overlay mode is required.
    */
   private void replaceWorker(Text       text,
                              NumberType pos,
                              NumberType len,
                              boolean    overlay)
   throws ErrorConditionException
   {
      // quick out
      if (this.isUnknown())
      {
         // nothing to do, you can't replace chars in the unknown value
         return;
      }
      
      // are any of our inputs unknown?
      if (text.isUnknown() || pos.isUnknown() || len.isUnknown())
      {
         // set our result to unknown and return
         setUnknown();
         return;
      }
      
      String rtxt = text.getValue();
      int    rpos = pos.intValue();
      int    rlen = len.intValue();
      
      // argument check
      if (rpos < 1)
      {
         TextOps.genNegativeIndexError();
         return;
      }
      
      // convert from 1-based to 0-based index
      rpos--;
      
      // -1 is equivalent of the length of the replacement text
      if (rlen == -1)
      {
         rlen = rtxt.length();
      }
      
      // replacement text is exactly as passed-in if not in overlay mode
      if (overlay)
      {
         // overlay mode means that the len parameter also changes the 
         // replacement text
         if (rtxt.length() < rlen)
         {
            // pad to the requested length
            rtxt = StringHelper.leftAlignText(rtxt, rlen);
         }
         else
         {
            // truncate the replacement text
            rtxt = rtxt.substring(0, rlen); 
         } 
      }
      
      StringBuffer sb = new StringBuffer();
      
      // start with the beginning of the string
      if (rpos >= value.length())
      {
         // pad to the requested position
         sb.append(StringHelper.leftAlignText(value, rpos));
      }
      else
      {
         // just substring the first portion
         sb.append(value.substring(0, rpos));
      }
      
      // insert our replacement text
      sb.append(rtxt);
      
      // index of the end of the string
      int elen = rpos + rlen;
      
      // if there is any text at the end of the string, add that
      if (value.length() > elen) 
      {
         sb.append(value.substring(elen));
      }
      
      // save the results into our instance data
      value = sb.toString();
   }

   /**
    * Helper to reset the associated instance's case sensitivity to a prior
    * value based on a notification that a particular scope has exited.  This
    * scope notification support is provided by 
    * {@link TransactionManager#registerFinalizable} and the {@link
    * Finalizable} interface.
    */
   private class SensitivityResetter
   implements Finalizable
   {
      /** The sensitivity to which the containing instance will be reset. */ 
      boolean oldCaseSens;
      
      /**
       * Construct an instance which can reset the containing instance's
       * case sensitivity when a notification is received.  See 
       * {@link character#setTemporaryCaseSensitive}.
       *
       * @param    oldCaseSens
       *           The case sensitivity to which the containing instance must
       *           be reset.
       */
      SensitivityResetter(boolean oldCaseSens)
      {
         this.oldCaseSens = oldCaseSens;
      }
      
      /**
       * Processes the end of scope notification and resets the associated 
       * instance's case sensitivity.  This notification is provided by the
       * {@link TransactionManager}.
       */
      public void finished()
      {
         Text.this.caseSens = oldCaseSens;
      }
      
      /**
       * Processes the iterate notification provided by the {@link
       * TransactionManager}.  This implementation does nothing.
       */
      public void iterate()
      {
      }
      
      /**
       * Processes the retry notification provided by the {@link
       * TransactionManager}.  This implementation does nothing.
       */
      public void retry()
      {
      }      
   }
}